from __future__ import annotations

import os
import re
from typing import Callable, ContextManager, TypeVar

from loguru import logger
from sqlalchemy import create_engine, text
from sqlalchemy.orm import DeclarativeBase, Session, scoped_session, sessionmaker
from sqlalchemy_utils import create_database, database_exists

from petisco.base.domain.persistence.database import Database
from petisco.extra.sqlalchemy.sql.mysql.mysql_connection import MySqlConnection
from petisco.extra.sqlalchemy.sql.sql_base import SqlBase
from petisco.extra.sqlalchemy.sql.sql_session_scope_provider import (
    sql_session_scope_provider,
)
from petisco.extra.sqlalchemy.sql.sqlite.sqlite_connection import SqliteConnection

T = TypeVar("T")

SqlSessionScope = Callable[..., ContextManager[Session]]


class SqlDatabase(Database[Session]):

    session_factory: sessionmaker | None = None

    def __init__(
        self,
        connection: SqliteConnection | MySqlConnection,
        *,
        alias: str | None = None,
        print_sql_statements: bool = False,
        use_scoped_session: bool = True,
        initial_statements_filename: str = None,
    ):
        self.connection = connection
        self.print_sql_statements = print_sql_statements
        self.use_scoped_session = use_scoped_session
        self.initial_statements_filename = initial_statements_filename
        self._check_connection()
        super().__init__(alias)

    def _check_connection(self) -> None:
        if not self.connection or not isinstance(
            self.connection, (SqliteConnection, MySqlConnection)
        ):
            raise ConnectionError(
                "SqlDatabase needs a valid SqliteConnection or MySqlConnection connection"
            )

    def initialize(self, base: DeclarativeBase = SqlBase) -> None:
        engine = create_engine(
            self.connection.url,
            json_serializer=lambda obj: obj,
            json_deserializer=lambda obj: obj,
            echo=self.print_sql_statements,
        )

        if not database_exists(engine.url):
            create_database(engine.url)
            base.metadata.create_all(engine)
            self._run_initial_statements(engine)

        self.session_factory = sessionmaker(bind=engine)

    def _run_initial_statements(self, engine) -> None:
        if self.initial_statements_filename:
            try:
                file = open(self.initial_statements_filename)
                statements = re.split(r";\s*$", file.read(), flags=re.MULTILINE)
                with engine.connect() as conn:
                    for statement in statements:
                        if statement:
                            conn.execute(text(statement))
                    conn.commit()
            except Exception as exc:  # noqa
                raise RuntimeError(
                    f"Error loading the initial_statements_filename={self.initial_statements_filename}. {str(exc)}"
                )

    def delete(self):
        if isinstance(self.connection, SqliteConnection):
            if os.path.exists(self.connection.database_name):
                os.remove(self.connection.database_name)
        else:
            logger.warning(
                "SqlDatabase do not implement delete to mitigate possible problems in production"
            )

    def clear_data(self, base: DeclarativeBase = SqlBase) -> None:
        if isinstance(self.connection, SqliteConnection):
            session_scope = self.get_session_scope()
            with session_scope() as session:
                for table in reversed(base.metadata.sorted_tables):
                    session.execute(table.delete())
        else:
            logger.warning(
                "SqlDatabase do not implement clear_data to mitigate possible problems in production"
            )

    def get_session_scope(self) -> Callable[..., ContextManager[Session]]:
        if self.session_factory is None:
            raise RuntimeError(
                "SqlDatabase must run initialize() before get_session_scope()"
            )

        if self.use_scoped_session:
            Session: Callable[..., Session] = scoped_session(
                self.session_factory
            )  # noqa
        else:
            Session: Callable[..., Session] = self.session_factory  # noqa
        return sql_session_scope_provider(Session)

    def is_available(self):
        try:
            context = self.get_session_scope()
            with context() as session:
                session.execute(text("SELECT 1"))
                _is_available = True
        except Exception:  # noqa E722
            _is_available = False
        return _is_available



from __future__ import annotations

from typing import Any, Callable

from meiga import AnyResult, Result


def unwrap_result_handler(result: Result) -> Any:
    assert isinstance(result, Result), "result_handler input must be a Result"
    return result.unwrap()


def custom_message_handler(
    message: dict[str, Any]
) -> Callable[[AnyResult], dict[str, Any]]:
    def _result_handler(_: AnyResult) -> dict[str, Any]:
        return message

    return _result_handler



from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Union

from loguru import logger
from pydantic import Field
from pydantic_settings import BaseSettings

from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.controller.error_map import ErrorMap
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.extra.rabbitmq.dependencies import get_default_message_dependencies
from petisco.extra.slack.dependencies import get_default_notifier_dependencies


class Application(BaseSettings):
    name: str
    version: str
    organization: str
    deployed_at: datetime
    environment: str = Field(default="local", env="ENVIRONMENT")
    dependencies_provider: Callable[..., List[Dependency]] = lambda: []
    configurers: List[ApplicationConfigurer] = []
    shared_error_map: Union[ErrorMap, None] = Field(default=dict())

    def __init__(self, **data: Any) -> None:
        info = ApplicationInfo(
            name=data["name"],
            organization=data["organization"],
            version=data["version"],
            deployed_at=data.get("deployed_at"),
            force_recreation=True,
            shared_error_map=data.get("shared_error_map", dict()),
        )
        deployed_at = info.deployed_at.strftime("%m/%d/%Y, %H:%M:%S")
        logger.info(
            f"Application: {info.name} {info.version} ({info.organization}) deployed at {deployed_at}"
        )
        super().__init__(**data)

    def configure(
        self, testing: bool = False, overwrite_dependencies: bool = False
    ) -> None:

        before_dependencies_configurers = [
            configurer
            for configurer in self.configurers
            if configurer.execute_after_dependencies is False
        ]
        name_configurers = [
            configurer.__class__.__name__
            for configurer in before_dependencies_configurers
        ]
        logger.info(
            f"Application: running configurators before setting dependencies {name_configurers}..."
        )
        for configurer in before_dependencies_configurers:
            configurer.execute(testing)

        logger.info("Application: setting dependencies...")
        Container.set_dependencies(self.get_dependencies(), overwrite_dependencies)

        after_dependencies_configurers = [
            configurer
            for configurer in self.configurers
            if configurer.execute_after_dependencies is True
        ]

        name_configurers = [
            configurer.__class__.__name__
            for configurer in after_dependencies_configurers
        ]
        logger.info(
            f"Application: running configurators after setting dependencies {name_configurers}..."
        )
        for configurer in after_dependencies_configurers:
            configurer.execute(testing)

        logger.info("Application: successful configuration")

    def get_dependencies(self) -> List[Dependency]:
        # TODO: review default dependencies in v2

        default_dependencies = (
            get_default_message_dependencies() + get_default_notifier_dependencies()
        )
        default_dependencies_dict = {
            dependency.get_key(): dependency for dependency in default_dependencies
        }
        provided_dependencies = self.dependencies_provider()
        provided_dependencies_dict = {
            dependency.get_key(): dependency for dependency in provided_dependencies
        }
        # This merged_dependencies will give preference to provided_dependencies_dict over default_dependencies_dict
        merged_dependencies = {
            **default_dependencies_dict,
            **provided_dependencies_dict,
        }
        return list(merged_dependencies.values())

    def clear(self) -> None:
        Container.clear()

    def info(self) -> Dict[str, Any]:
        info = self.dict()
        info["deployed_at"] = self.deployed_at.strftime("%m/%d/%Y, %H:%M:%S")
        info["dependencies"] = {
            dependency.type.__name__: dependency.get_instance().info()
            for dependency in self.get_dependencies()
        }
        del info["dependencies_provider"]
        del info["configurers"]
        del info["shared_error_map"]
        return info

    def was_deploy_few_minutes_ago(self, minutes: int = 25) -> bool:
        return datetime.utcnow() < self.deployed_at + timedelta(minutes=minutes)

    def publish_domain_event(self, domain_event: DomainEvent) -> None:
        try:
            domain_event_bus = Container.get(DomainEventBus)
            domain_event_bus.publish(domain_event)
        except Exception as exc:  # noqa
            raise TypeError(
                f"To publish an event to the domain event bus, please add a dependency with type `DomainEventBus` on Application dependencies. {str(exc)}"
            )

    def notify(self, message: NotifierMessage) -> None:
        try:
            notifier = Container.get(Notifier)
            notifier.publish(message)
        except:  # noqa
            raise TypeError(
                'To notify the deploy, please add a dependency with name "notifier" on Application dependencies'
            )



from __future__ import annotations

from pydantic.main import BaseModel

from petisco.base.domain.message.message import Message


class MessageSubscriberInfo(BaseModel):
    message_name: str
    message_version: int
    message_type: str

    @staticmethod
    def from_class_type(class_type: type[Message]) -> MessageSubscriberInfo:
        message = class_type()
        return MessageSubscriberInfo(
            message_name=message.get_message_name(),
            message_version=message.get_message_version(),
            message_type=message.get_message_type(),
        )



from __future__ import annotations

from collections import defaultdict
from typing import Generic, TypeVar

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.misc.singleton import Singleton

T = TypeVar("T")


class Container(Generic[T], metaclass=Singleton):
    """
    Singleton which contains set dependencies (List[Dependency)) prepared to be instantiated in order to be injected
     in the UseCases of our application
    """

    def __init__(self) -> None:
        self.dependencies: dict[str, Dependency] = defaultdict()

    @staticmethod
    def get(base_type: type[T], *, alias: str | None = None) -> T:
        """
        Returns an instance of set Dependency.
        """
        container = Container()

        key = base_type if not alias else alias
        dependency = container.dependencies.get(key)
        if dependency is None:
            key_str = key.__name__ if isinstance(key, type) else key
            raise IndexError(
                f"Invalid dependency. `{key_str}` is not found within available dependencies [{container.get_available_dependencies()}]"
            )
        instance = dependency.get_instance()
        return instance

    @staticmethod
    def set_dependencies(
        dependencies: list[Dependency] | None = None, overwrite: bool = False
    ) -> None:
        """
        Set dependencies from a list of them.
        """
        if dependencies is None:
            dependencies = []
        Container()._set_dependencies(dependencies, overwrite)

    @staticmethod
    def get_available_dependencies() -> list[str]:
        """
        Returns the names (keys) of set dependencies.
        """
        keys = list(Container().dependencies.keys())
        return [key if isinstance(key, str) else key.__name__ for key in keys]

    def _set_dependencies(
        self, input_dependencies: list[Dependency], overwrite: bool = False
    ) -> None:
        for dependency in input_dependencies:
            if dependency.alias:
                if dependency.alias in self.dependencies and not overwrite:
                    raise IndexError(
                        f"Container: dependency (alias={dependency.alias}) is already added to dependencies. Check "
                        f"set_dependencies input "
                    )
                self.dependencies[dependency.alias] = dependency
            else:
                if dependency.type:
                    if dependency.type in self.dependencies and not overwrite:
                        raise IndexError(
                            f"Container: dependency (type={dependency.type.__name__}) is already added to "
                            f"dependencies. Use Dependency alias to set different dependencies with the same base type "
                        )
                    self.dependencies[dependency.type] = dependency



from __future__ import annotations

from abc import abstractmethod

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message_bus import MessageBus


class DomainEventBus(MessageBus[DomainEvent]):
    """
    A base class to implement an infrastructure-based bus to publish events.
    """

    @abstractmethod
    def publish(self, domain_event: DomainEvent | list[DomainEvent]) -> None:
        """
        Publish a DomainEvent or a list of DomainEvents
        """
        raise NotImplementedError

    def _check_is_domain_event(
        self, domain_event: DomainEvent | list[DomainEvent]
    ) -> None:
        if not domain_event or not issubclass(domain_event.__class__, DomainEvent):
            raise TypeError(
                f"{self.__class__.__name__} only publishes DomainEvent objects"
            )

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError



from petisco.extra.elastic.is_elastic_available import is_elastic_available

elastic = list()
if is_elastic_available():
    from petisco.extra.elastic.async_elastic_database import (
        AsyncElasticDatabase,
        AsyncElasticSessionScope,
    )
    from petisco.extra.elastic.elastic_connection import ElasticConnection
    from petisco.extra.elastic.elastic_database import (
        ElasticDatabase,
        ElasticSessionScope,
    )
    from petisco.extra.elastic.legacy_elastic_database import LegacyElasticDatabase

    elastic = [
        "ElasticConnection",
        "LegacyElasticDatabase",
        "ElasticDatabase",
        "AsyncElasticDatabase",
        "ElasticSessionScope",
        "AsyncElasticSessionScope",
    ]
else:
    elastic = []

__all__ = elastic



from __future__ import annotations

from typing import Any, Callable, ContextManager

from petisco.base.domain.persistence.database import Database, T


class FakeDatabase(Database):
    def initialize(self, *args: Any, **kwargs: Any) -> None:
        pass

    def delete(self) -> None:
        pass

    def clear_data(self) -> None:
        pass

    def is_available(self) -> bool:
        pass

    def get_session_scope(self) -> Callable[..., ContextManager[T]]:
        pass



__all__ = []

from petisco.extra.redis.is_redis_available import is_redis_available

if is_redis_available():

    from petisco.extra.redis.application.message.bus.redis_command_bus import (
        RedisCommandBus,
    )
    from petisco.extra.redis.application.message.bus.redis_domain_event_bus import (
        RedisDomainEventBus,
    )

    __all__ += [
        "RedisCommandBus",
        "RedisDomainEventBus",
    ]



from petisco.extra.sqlalchemy.is_sqlalchemy_available import is_sqlalchemy_available

if is_sqlalchemy_available():
    from petisco.extra.sqlalchemy.sql.async_sql_database import (
        AsyncSqlDatabase,
        AsyncSqlSessionScope,
    )
    from petisco.extra.sqlalchemy.sql.mysql.legacy_mysql_database import (
        LegacyMySqlDatabase,
    )
    from petisco.extra.sqlalchemy.sql.mysql.mysql_connection import MySqlConnection
    from petisco.extra.sqlalchemy.sql.sql_base import SqlBase
    from petisco.extra.sqlalchemy.sql.sql_database import SqlDatabase, SqlSessionScope
    from petisco.extra.sqlalchemy.sql.sql_executor import SqlExecutor
    from petisco.extra.sqlalchemy.sql.sqlite.legacy_sqlite_database import (
        LegacySqliteDatabase,
    )
    from petisco.extra.sqlalchemy.sql.sqlite.sqlite_connection import SqliteConnection

    __all__ = [
        "MySqlConnection",
        "LegacyMySqlDatabase",
        "SqliteConnection",
        "LegacySqliteDatabase",
        "SqlExecutor",
        "SqlDatabase",
        "AsyncSqlDatabase",
        "SqlSessionScope",
        "AsyncSqlSessionScope",
        "SqlBase",
    ]
else:
    __all__ = []



from __future__ import annotations

from abc import abstractmethod
from typing import Any, cast

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.controller.error_map import ErrorMap
from petisco.base.application.controller.meta_controller import MetaController
from petisco.base.misc.result_mapper import ResultMapper, default_failure_handler


class AsyncController(metaclass=MetaController):
    """
    A base class for creating async controllers.
    Inherit from this class to convert to domain the request values, configure middlewares and instantiate and execute
     a UseCase.
    """

    @staticmethod
    def get_default_mapper() -> ResultMapper:
        return ResultMapper()

    @staticmethod
    def get_config_mapper(config: dict[str, Any]) -> ResultMapper:
        return ResultMapper(
            error_map=cast(ErrorMap, getattr(config, "error_map", None)),
            success_handler=getattr(config, "success_handler", lambda result: result),
            failure_handler=getattr(config, "failure_handler", default_failure_handler),
        )

    @abstractmethod
    async def execute(
        self, *args: tuple[str, ...], **kwargs: dict[str, Any]
    ) -> AnyResult:
        return NotImplementedMethodError



from __future__ import annotations

import re
from typing import AsyncContextManager, Callable, TypeVar

from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, scoped_session
from sqlalchemy_utils import database_exists

from petisco.base.domain.persistence.async_database import AsyncDatabase
from petisco.extra.sqlalchemy.sql.async_sql_session_scope_provider import (
    async_sql_session_scope_provider,
)
from petisco.extra.sqlalchemy.sql.sql_base import SqlBase
from petisco.extra.sqlalchemy.sql.sql_database import SqlDatabase

T = TypeVar("T")

AsyncSqlSessionScope = Callable[..., AsyncContextManager[AsyncSession]]


class AsyncSqlDatabase(SqlDatabase, AsyncDatabase[AsyncSession]):

    async_session_factory: async_sessionmaker | None = None

    async def initialize(self, base: DeclarativeBase = SqlBase) -> None:
        engine = create_async_engine(
            self.connection.url,
            json_serializer=lambda obj: obj,
            json_deserializer=lambda obj: obj,
            echo=self.print_sql_statements,
        )

        if not database_exists(engine.url):
            # await create_database(engine.url)
            async with engine.begin() as conn:
                await conn.run_sync(base.metadata.create_all)
                await self._async_run_initial_statements(conn)

        self.async_session_factory = async_sessionmaker(bind=engine)

    async def _async_run_initial_statements(self, conn) -> None:
        if self.initial_statements_filename:
            try:
                file = open(self.initial_statements_filename)
                statements = re.split(r";\s*$", file.read(), flags=re.MULTILINE)
                for statement in statements:
                    if statement:
                        await conn.execute(text(statement))
            except Exception as exc:  # noqa
                raise RuntimeError(
                    f"Error loading the initial_statements_filename={self.initial_statements_filename}. {str(exc)}"
                )

    def get_session_scope(self) -> Callable[..., AsyncContextManager[AsyncSession]]:
        if self.async_session_factory is None:
            raise RuntimeError(
                "AsyncSqlDatabase must run initialize() before get_session_scope()"
            )

        if self.use_scoped_session:
            Session = scoped_session(self.async_session_factory)  # noqa
        else:
            Session = self.async_session_factory  # noqa
        return async_sql_session_scope_provider(Session)

    async def is_available(self):
        try:
            context = self.get_session_scope()
            async with context() as session:
                await session.execute(text("SELECT 1"))
                _is_available = True
        except Exception:  # noqa E722
            _is_available = False
        return _is_available
